#!/usr/bin/perl -w
# 

use strict;
use Cwd;
use File::Basename;
use Date::Calc;
use LWP::UserAgent;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

my %http_info;
my %files;

my $www_data_dir     = '/var/www/fci/';
my $client_data_dir  = 'client-data/';
my $eoc_data_dir     = 'eoc-data/';
my $xml_data_dir     = 'xml-data/';

my %cfg = ( self                     => 'fci-update-eoc-stats.pl',
            version                  => '0.2',
	    
	    verbose                  => 0,
	    quiet                    => 0,

	    'www-data-dir'           => $www_data_dir,
	    'client-data-dir'        => $client_data_dir,
	    'eoc-data-dir'           => $eoc_data_dir,
	    'xml-data-dir'           => $xml_data_dir,
	    'username-list-file'     => $xml_data_dir.'username-list.xml',
	    'eoc-stats-file'         => $xml_data_dir.'eoc-stats.xml',

	    'download'               => 1,
	    'sleep'                  => 2,
          );


# Get commandline parameters
Getopt::Mixed::init("dir=s
                     sleep=i
                     no-download
                     verbose
		     quiet
		     help
		     d>dir
		     s>sleep
		     n>no-download
		     v>verbose
		     q>quiet
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit 0;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{'sleep'}        = $value if($option eq 'sleep');
	$cfg{'download'}     = 0      if($option eq 'no-download');
	
	$cfg{verbose}    = 1      if($option eq 'verbose');
	$cfg{quiet}      = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# if quiet was requested together with verbose, disable verbose which contradicts quiet
if($cfg{quiet} && $cfg{verbose}) {
	#print "--quiet overrides --verbose\n";
	$cfg{verbose} = 0;
}

# Check the sleep time
if(!$cfg{sleep} || $cfg{sleep} < 1) {
	print "Warning: Too short sleep time specified, using 1 second instead.\n";
	$cfg{sleep} = 1;
}

# add directory separator to the end of the www-data-dir path if missing
$cfg{'www-data-dir'} .= '/' if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}


# show what's print()ed without waiting for the newline $|=1;
$|=1;


# check file existance and permissions
if(!-r $cfg{'username-list-file'}) {
	print "Error: Cannot read username-list.xml file!\n";
	exit 1;
}
if(!-r $cfg{'eoc-stats-file'}) {
	print "Error: Cannot read eoc-stats.xml file!\n";
	exit 1;
}


# read the xml files
my $username_list = read_xml(%cfg, Name => 'username-list') || die "Error: Cannot read username-list XML file!\n";
my $eoc_stats     = read_xml(%cfg, Name => 'eoc-stats')     || die "Error: Cannot read eoc-stats XML file!\n";

#print Dumper($username_list);

# create useragent object which will handle the HTTP connections
my $ua  = LWP::UserAgent->new();
   $ua->agent("$cfg{self}/$cfg{version}");

my $updated    = 0;
my $failed     = 0;
my $max_failed = 5;

# for each users listed download it's Extreme Overclocking XML file
foreach my $userid (keys %{$username_list->{username_list}->{username}}) {

	my $username = $username_list->{username_list}->{username}->{$userid}->{name};
	my $team     = $username_list->{username_list}->{username}->{$userid}->{team};
	
	# old: http://folding.extremeoverclocking.com/xml/user_summary.php?un=rph_iv&t=0
	# new: http://folding.extremeoverclocking.com/xml/8zb43r/user_summary.php?un=rph_iv&t=0
	# The new URL has been fased out again as of 2008-11-19
	my $url   = 'http://folding.extremeoverclocking.com/xml/user_summary.php?un=';
	   $url  .= $username;
	   $url  .= '&t=';
	   $url  .= $team;
	my $file  = $cfg{'eoc-data-dir'}.$username.'-'.$team.'.xml';
	my $xml   = '';
	
	if(!$cfg{download}) {
		$files{$userid} = $file;
	}
	else {
		if($xml = fetch_url($ua, $url)) {
			# continue to save the content
		}
		else {
			if($failed < $max_failed) {
				$failed++;

				# Failed to download the file, skip to the next username
				print "Skipping to next download.\n" if($cfg{verbose});
				sleep($cfg{sleep}); # don't check to many users per second
				next;
			}
			else {
				print "Error: Failed to download $failed summaries, terminating the program!\n";
				exit 1;
			}
		}
		
		# save file to disc
		if(save_file($file, $xml)) {
			# continue
			
			$files{$userid} = $file;

			$eoc_stats->{eoc_stats}->{username}->{$userid}->{name} = $username;
			$eoc_stats->{eoc_stats}->{username}->{$userid}->{team} = $team;
			$eoc_stats->{eoc_stats}->{username}->{$userid}->{url}  = $url;
			$eoc_stats->{eoc_stats}->{username}->{$userid}->{file} = $file;

			$updated++;
		}
		else {
			print "Error: Could not save file! ($file)\n";
			exit 1;
		}
	}

	print "\n" if($cfg{verbose});
	sleep($cfg{sleep}); # don't check to many users per second
}

if($updated > 0) {
	print "Saving eoc-stats XML file ...\n" if($cfg{verbose});
	my ($success, $reason) = write_xml( %cfg,
					    Name => 'eoc-stats',
					    XML  => $eoc_stats
					  );
		
	if(!$success) {
		print $reason;
		exit 1;
	}
	else {
		if($cfg{verbose}) {
			print "Successfully saved eoc-stats XML file\n";
		}
	}
}

exit 0; # exit normally

################################[ Subroutines ]################################

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}


##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file    = $_[0];
	my $verbose = $_[1] ? $_[1] : 0;
	
	print "Reading file from disk: $file ... " if($verbose);
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($verbose);

		return $data;
	}
	else {
		print "FAILED!\n" if($verbose);
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode    => file premission in octal
# 	         binary  => flag to set binary mode (only needed on Windows)
# 	         verbose => flag to set verbose output
#     Returns: 1 for successfull save or undef for unsuccessfull save
#              the reason for the failure or an empty string
# Description: write a file to disk if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	my ($success, $reason) = (1, '');

	print "Saving file to disk: $filename ... " if($option{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($option{verbose});
		
		open(F, "> $filename") || return (0, "Error: Cannot save file: $filename ($!)\n"); 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($option{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($option{verbose});
			chmod oct($option{mode}), $filename;
		}
		else {
			if($option{verbose}) {
				print "Ignoring invalid mode: ";
				print $option{mode};
				print "\n";
			}
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($option{verbose});
		return (0, "Error: File does not exists or is unwriteable!\n");
	}
}


#############
#
#     Routine: fetch_url()
#  Parameters: user-agent object
#              url
#     Returns: page content on success or
#              undef in case of no success
# Description: retrieve a file via an HTTP GET request
# 

sub fetch_url {
	my $ua  = $_[0];
	my $url = $_[1];
	
	my $req = HTTP::Request->new(GET => $url);
	
	print "Retrieving URL: $url ... " if($cfg{verbose});
	my $res = $ua->request($req);
	if($res->is_success) {
		print "Success\n" if($cfg{verbose});
		
		return $res->content;
	}
	else {
		print "Failed!\n" if($cfg{verbose});
		print "Error: Failed to retrieve URL! ($url)\n";
		print "HTTP Status: ".$res->code." ".$res->message."\n";
			
		return undef;
	}
}
																	

##############
#
#     Routine: valid_url()
#  Parameters: url
#     Returns: 1 on success or
#              2 in case of authentication required or
#              0 otherwise
# Description: verify the access to a file via a HTTP HEAD request and
#              detect requirement of HTTP Authentication
# 

sub valid_url {
	my $url = $_[0];
	
	my $ua  = LWP::UserAgent->new();
	   $ua->agent($cfg{useragent});
	
	my $req = HTTP::Request->new(HEAD => $url);
	my $res = $ua->request($req);
	
	# store realm and http status info in global hash
	$http_info{$url}{realm}  = $res->header('WWW-Authenticate');
	$http_info{$url}{status} = $res->code." ".$res->message;

	if($res->is_success()) {
		return 1;
	}
	elsif($res->code == 401) {
		return 2;
	}
	else {
		return 0;
	}
}


##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats|eoc-user-summary)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'eoc-stats') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'eoc-user-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'eoc-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'eoc-user-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
		
		# allow group users write permissions
		chmod(0664, $args{File});
	}

	return ($success, $reason);
}

sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>         Path of the FCI server directory\n";
	print "-n, --no-download       Don't download any files, use the ones on disk\n";
	print "-v, --verbose           Enable verbose output\n";
	print "-q, --quiet             Don't print any output at all, except for errors\n";
	print "-h, --help              Show this usage information\n";
}
